import unittest
import numpy as np
import surfa as sf


class test_affine(unittest.TestCase):

    def test_init(self):
        """
        TODOC
        """

        # TODO test init with source/target images and surfaces

        for ndim in (2, 3):
            eye = np.eye(ndim + 1)
            aff = sf.Affine(eye)
            self.assertTrue(aff.ndim == ndim)
            
            aff = sf.Affine(eye[:ndim, :], transform_space='image-to-world')
            self.assertTrue(aff.ndim == ndim)
            
            aff = sf.transform.identity(ndim=ndim)
            self.assertTrue(np.array_equal(aff.matrix, eye))

            aff = sf.transform.random_affine(ndim=ndim)

        # bad entries
        self.assertRaises(Exception, sf.Affine, np.eye(2))
        self.assertRaises(Exception, sf.Affine, np.eye(5))

    def test_equivalence(self):
        """
        Test that affine equivalence functions are doing the right thing. The combinations of space
        and geometry are nonsensical here, but it doesn't matter.
        """
        
        for ndim in (2, 3):

            source = sf.ImageGeometry(np.repeat(128, ndim))
            target = sf.ImageGeometry(np.repeat(128, ndim))

            a = sf.transform.random_affine(ndim=ndim, source=source, target=target, transform_space='image-to-world')
            b = sf.Affine(a.matrix.copy(), source=source, target=target, transform_space='image-to-world')
            self.assertTrue(sf.transform.affine_equal(a, b))

            b = sf.Affine(a.matrix.copy(), source=target, target=source, transform_space='image-to-world')
            # self.assertFalse(sf.transform.affine_equal(a, b))  # todo
            self.assertTrue(sf.transform.affine_equal(a, b, matrix_only=True))

            b = sf.Affine(a.matrix.copy(), source=source, target=target, transform_space='surface-to-world')
            self.assertFalse(sf.transform.affine_equal(a, b))

            b = sf.Affine(a.matrix.copy() + 1e-5, source=source, target=target, transform_space='image-to-world')
            self.assertFalse(sf.transform.affine_equal(a, b, tol=0.9e-5))
            self.assertTrue(sf.transform.affine_equal(a, b, tol=1.1e-5))

    def test_point_transform(self):
        """
        todoc
        """
        # todo
        for ndim in (2, 3):

            affine = sf.transform.identity(ndim=ndim)

            moving = np.random.randn(ndim)
            moved = affine.transform(moving)
            # todo test difference

            moving = np.random.randn(5, ndim)
            moved = affine.transform(moving)
            # todo test diffference

            # TODO should we even support this?
            # moving = np.random.randn(5, 5, ndim)
            # moved = affine.transform(moving)
            # todo test diffference

            moving = np.random.randn(ndim + 1)
            self.assertRaises(Exception, affine.transform, moving)
            moving = np.random.randn(5, ndim + 1)
            self.assertRaises(Exception, affine.transform, moving)

            translation = np.random.randn(ndim)
            affine = sf.transform.compose_affine(translation=translation, ndim=ndim)
            moving = np.random.randn(ndim)
            moved = affine.transform(moving)
            true = moving + translation
            # todo add X and test differente

    def test_inverse(self):
        """
        Test affine matrix inversion. Important to test that source and target information is swapped
        in the resulting affine. The combinations of space and geometry are nonsensical here, but
        it doesn't matter.
        """

        for ndim in (2, 3):

            source = sf.ImageGeometry(np.repeat(128, ndim))
            target = sf.ImageGeometry(np.repeat(128, ndim))

            aff = sf.transform.identity(ndim, source=source, target=target, transform_space='image-to-world')
            inv = aff.inverse()

            self.assertTrue(aff.source is inv.target)
            self.assertTrue(aff.target is inv.source)
            self.assertTrue(aff.transform_space.source == inv.transform_space.target)
            self.assertTrue(aff.transform_space.target == inv.transform_space.source)
            self.assertTrue(np.array_equal(inv.matrix, np.linalg.inv(aff.matrix)))

    def test_matmul(self):
        """
        Test matrix multiplication of the affines. The important thing here is to make sure
        the source and target information is taken from the correct affines. The combinations
        of space and geometry are nonsensical here, but it doesn't matter.
        """
        for ndim in (2, 3):
            source = sf.ImageGeometry(np.repeat(128, ndim))
            target = sf.ImageGeometry(np.repeat(128, ndim))
            intermediate = sf.ImageGeometry(np.repeat(128, ndim))
            a = sf.transform.identity(ndim, source=intermediate, target=target, transform_space='world-to-surface')
            b = sf.transform.identity(ndim, source=source, target=intermediate, transform_space='image-to-world')
            c = a @ b
            self.assertTrue(sf.transform.geometry.image_geometry_equal(c.source, source))
            self.assertTrue(sf.transform.geometry.image_geometry_equal(c.target, target))
            self.assertTrue(c.transform_space == 'image-to-surface')

    def test_composition(self):
        """
        TODOC
        """

        for ndim in (2, 3):

            # identity (default) affine
            aff = sf.transform.compose_affine(ndim=ndim)
            self.assertTrue(np.array_equal(np.eye(ndim + 1), aff.matrix))
            translation, rotation, scale, shear = aff.decompose()
            self.assertTrue(np.all(translation == 0.0))
            self.assertTrue(np.all(rotation == 0.0))
            self.assertTrue(np.all(scale == 1.0))
            self.assertTrue(np.all(shear == 0.0))

            # random affine
            translation = np.random.uniform(-20, 20, ndim)
            rotation = np.random.uniform(-90, 90, 1 if ndim == 2 else 3)
            scale = np.random.uniform(0.5, 2.0, ndim)
            shear = np.random.uniform(-0.1, 0.1, 1 if ndim == 2 else 3)
            aff = sf.transform.compose_affine(translation, rotation, scale, shear, ndim=ndim)
            n_translation, n_rotation, n_scale, n_shear = aff.decompose()
            self.assertTrue(np.allclose(translation, n_translation))
            self.assertTrue(np.allclose(rotation, n_rotation))
            self.assertTrue(np.allclose(scale, n_scale))
            self.assertTrue(np.allclose(shear, n_shear))

            # radians
            rotation = np.radians(rotation)
            aff_radians = sf.transform.compose_affine(translation, rotation, scale, shear, ndim=ndim, degrees=False)
            self.assertTrue(np.array_equal(aff.matrix, aff_radians.matrix))
            n_translation, n_rotation, n_scale, n_shear = aff.decompose(degrees=False)
            self.assertTrue(np.allclose(translation, n_translation))
            self.assertTrue(np.allclose(rotation, n_rotation))
            self.assertTrue(np.allclose(scale, n_scale))
            self.assertTrue(np.allclose(shear, n_shear))


class test_space(unittest.TestCase):

    def test_init(self):
        """
        TODOC
        """

        # include some random capitalization
        good_names = [
            'i', 'image', 'IMG', 'vox', 'voxel',
            'W', 'RAS', 'world',
            'S', 'surf', 'surfACE',
        ]
        for name in good_names:
            space = sf.Space(name)

        # an unknown name
        self.assertRaises(Exception, sf.Space, None)
        self.assertRaises(Exception, sf.Space, 'bad')

    def test_equivalence(self):
        """
        TODOC
        """
        self.assertTrue(sf.Space('image') == sf.Space('voxel'))
        self.assertTrue(sf.Space('surface') == 's')
        self.assertTrue(sf.Space('world') == 'ras')
        self.assertFalse(sf.Space('world') == 'image')
        self.assertFalse(sf.Space('world') == None)


class test_image_geometry(unittest.TestCase):

    def test_init(self):
        """
        TODO
        """
        pass

    def test_update(self):
        """
        TODO
        """
        pass

    def test_writable(self):
        """
        TODO
        """
        pass

    def test_copy(self):
        """
        TODO
        """
        pass


class test_orientation(unittest.TestCase):

    def test_init(self):
        """
        TODO
        """
        pass


class test_deformation(unittest.TestCase):

    def test_init(self):
        """
        TODO
        """
        pass


if __name__ == '__main__':
    unittest.main()



import numpy as np


def sane_slicing(shape, index_expression):
    """
    Clean up an index expression such that the result is a tuple with
    the proper number of dimensions and slicings to match a target shape.

    Parameters
    ----------
    shape : tuple of int
        Target array shape.
    index_expression : tuple
        Numpy-style index expression.

    Returns
    -------
    index_expression : tuple
        Cleaned index expression.
    """
    ndim = len(shape)
    slicing = [None] * len(shape)
    index_expression = np.index_exp[index_expression]

    def make_sane_dimension(x, length, axis):
        if isinstance(x, slice):
            return slice(*x.indices(length))
        if isinstance(x, int):
            if x < 0:
                if x < -length:
                    raise IndexError(f'index {x} is out of bounds for axis {axis} with size {length}')
                x = length + x
            elif x >= length:
                raise IndexError(f'index {x} is out of bounds for axis {axis} with size {length}')
            return x
        else:
            raise IndexError('only integers, slices (`:`), and ellipsis (`...`) are valid indices')

    for i, x in enumerate(index_expression):

        if x is not Ellipsis:
            slicing[i] = make_sane_dimension(x, shape[i], i)
            continue

        for i, x in enumerate(reversed(index_expression[i + 1:])):
            if x is Ellipsis:
                raise IndexError('an index can only have a single ellipsis (`...`)')
            ni = i + 1
            slicing[-ni] = make_sane_dimension(x, shape[-ni], ndim - i)

        for i in range(ndim):
            if slicing[i] is None:
                slicing[i] = slice(*slice(None).indices(shape[i]))
        break

    return tuple(slicing)


def slicing_parameters(index_expression):
    """
    Convert a slicing index expression to a tuple of start and stop coordinates.
    This assumes the expression has been cleaned with `sane_slicing()`.

    Parameters
    ----------
    index_expression : tuple
        Numpy-style index expression.

    Returns
    -------
    tuple of int
        tuple of (start, stop) coordinates represented by the slicing.
    """
    start = []
    step = []
    for x in index_expression:
        if isinstance(x, slice):
            start.append(x.start)
            step.append(x.step)
        elif isinstance(x, int):
            start.append(x)
            step.append(1)
        else:
            raise ValueError('incompatible index expression `%s` - ensure that slicing is sane' % type(x))
    return (start, step)



#!/usr/bin/env python

import re
import pathlib
import packaging.version

from setuptools import setup
from setuptools import dist
from setuptools.extension import Extension


dist.Distribution().fetch_build_eggs(['cython', 'numpy'])


import numpy as np
from Cython.Build import cythonize


requirements = [
    'cython',
    'numpy',
    'scipy',
    'Pillow',
    'nibabel',
]

packages = [
    'surfa',
    'surfa.core',
    'surfa.transform',
    'surfa.image',
    'surfa.mesh',
    'surfa.io',
    'surfa.vis',
]

# base source directory
base_dir = pathlib.Path(__file__).parent.resolve()

# build cython modules
ext_modules = cythonize([
        Extension('surfa.image.interp', ['surfa/image/interp.pyx']),
    ],
    compiler_directives={'language_level' : '3'})

# extract the current version
init_file = base_dir.joinpath('surfa/__init__.py')
init_text = open(init_file, 'rt').read()
pattern = r"^__version__ = ['\"]([^'\"]*)['\"]"
match = re.search(pattern, init_text, re.M)
if not match:
    raise RuntimeError(f'Unable to find __version__ in {init_file}.')
version = match.group(1)
if isinstance(packaging.version.parse(version), packaging.version.LegacyVersion):
    raise RuntimeError(f'Invalid version string {version}.')

long_description = '''Surfa is a collection of Python (3.5+) utilities for medical image
analysis and mesh-based surface processing. It provides tools that operate on 3D image
arrays and triangular meshes with consideration of their representation in a world (or
scanner) coordinate system. While broad in scope, surfa is developed with particular
emphasis on neuroimaging applications, as it is an extension of the FreeSurfer brain
analysis software suite.
'''

# run setup
setup(
    name='surfa',
    version=version,
    description='Utilities for medical image processing and surface reconstruction.',
    long_description=long_description,
    author='Andrew Hoopes',
    author_email='freesurfer@nmr.mgh.harvard.edu',
    url='https://github.com/freesurfer/surfa',
    python_requires='>=3.5',
    packages=packages,
    ext_modules=ext_modules,
    include_dirs=[np.get_include()],
    package_data={'': ['*.pyx']},
    install_requires=requirements,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering',
    ],
)



from .space import Space
from .space import cast_space

from .affine import Affine
from .affine import cast_affine
from .affine import affine_equal
from .affine import identity
from .affine import compose_affine
from .affine import random_affine

from .geometry import ImageGeometry
from .geometry import cast_image_geometry
from .geometry import image_geometry_equal



from copy import deepcopy
import numpy as np

from surfa.core.array import conform_ndim
from surfa.core.labels import LabelLookup


class FramedArray:

    def __init__(self, basedim, data, labels=None, metadata=None):
        """
        Abstract class defining an ND array with data frames and additional meta information. This is
        the base type for volumes, slices, and overlays, which represent 3D, 2D, and 1D objects,
        respectively. This class should only be used as an internal base class and never be initialized
        directly.

        A `FramedArray` contains two distinct properties: buffer data and associated metadata. The data
        is always represented internally as a numpy array with an explicit number of dimensions. ND arrays
        can be optionally stacked along the last dimension, which represents individual 'data frames'. For
        example, a single-frame 3D `FramedArray` (defined as a `Volume`) might be of shape `(64, 64, 64)`,
        while a multi-frame `Volume` with the same 'base shape' might be `(64, 64, 64, 3)`. The frame axis
        is designed to represent a non-spatial dimension.

        The internal data buffer is wrapped, such that the `FramedArray` can be treated much like a numpy
        array. It can be manipulated with standard math and assignment operators, and it will be automatically
        converted to a numpy ndarray object if necessary.

        The input data is not copied, and the array should have ndims equal to the subclass' basedim (or
        basedim + 1). Any extra dimension is assumed to represent data frames.

        Parameters
        ----------
        basedim : int
            Array to pad.
        data : array_like
            Internal data array.
        metadata : dict
            Dictionary containing arbitrary array metadata.
        """

        # ensure abstract class isn't being used directly
        if not isinstance(basedim, int):
            raise TypeError('FramedArray cannot be initialized without setting a valid basedim')

        # set internal base dimension
        self._basedim = basedim

        # set data array
        self.data = data

        # initialize and set the private metadata dictionary
        self._metadata = {}
        self.metadata = metadata
        self.labels = labels

    def new(self, data):
        """
        Return a new instance of the array with updated data. Metadata is preserved.
        """
        return self.__class__(data=data, metadata=self.metadata)

    def copy(self):
        """
        Return a deep copy of the object.
        """
        return deepcopy(self)

    def zeros(self, dtype=None, frames=None, order='K'):
        """
        Return a copy of the framed array with all elements set to zero.

        Parameters
        ----------
        dtype : np.dtype
            Data type of new array.
        frames : int
            Number of frames to allocate. If `None`, the number of frames is
            determined from the source array.
        order : {‘C’, ‘F’, ‘K’}
            Controls the memory layout order of the result. ‘C’ means C order, ‘F’ means
            Fortran order, and ‘K’ means as close to the order the array elements appear
            in memory as possible.

        Returns
        -------
        FramedArray
            Array with zero values.
        """
        nframes = frames if frames is not None else self.nframes
        shape = (*self.baseshape, nframes)
        dtype = dtype if dtype is not None else self.dtype
        if order == 'K':
            order = 'F' if self.data.flags.f_contiguous else 'C'
        return self.new(np.zeros(shape, dtype=dtype, order=order))

    def __repr__(self):
        """
        Print out some basic information regarding shape and dtype. Should keep it simple.
        """
        return f'sf.{self.__class__.__name__}(shape={self.shape}, dtype={self.dtype})'

    @property
    def metadata(self):
        """
        Metadata dictionary.
        """
        return self._metadata

    @metadata.setter
    def metadata(self, value):
        """
        Replace the metadata dictionary. Will always make a deep copy of the new dictionary.
        """
        self._metadata = deepcopy(value) if value is not None else {}

    @property
    def labels(self):
        """
        Label lookup for segmentation data.
        """
        return self._metadata.get('labels')

    @labels.setter
    def labels(self, value):
        if value is None:
            self._metadata.pop('labels', None)
        elif not isinstance(value, LabelLookup):
            raise ValueError(f'labels expected LabelLookup object, but got object of type {value.__class__.__name__}')
        else:
            self._metadata['labels'] = value.copy()

    @property
    def basedim(self):
        """
        Base dimensionality of the array (excludes frame dimension).
        """
        return self._basedim

    @property
    def data(self):
        """
        Core data numpy ndarray.
        """
        return self._data

    @data.setter
    def data(self, value):
        """
        Setter for the private data array. Ensures dimensionality is correct. If the base-shape of
        the data has changed after updating, the private `_shape_changed()` hook will be called.
        """

        # make sure a string (filename) isn't being provided for input data - a common mistake
        if isinstance(value, str):
            raise TypeError('unexpected string for `data` parameter. Expected a {basedim}D array')

        # existing arrays are not copied
        value = np.asarray(value)

        # run a few sanity checks on the input data shape
        if value.ndim < 1:
            raise ValueError('array data cannot be set to scalar')

        # instead of throwing an error, data with fewer dimensions than
        # expected should be reshaped with added axes
        if value.ndim < self._basedim:
            value = conform_ndim(value, self._basedim)

        # single-framed arrays will always be represented by an array with
        # dimensionality equivalent to basedim
        if value.ndim == (self._basedim + 1) and value.shape[-1] == 1:
            value = value.squeeze(axis=-1)

        # throw an error if input array has more dimensions than the framed base
        if value.ndim > self._basedim + 1:
            raise ValueError(f'array data cannot be set from data with {value.ndim} dims')

        # check for shape changes, so we can update geometry if necessary
        current = getattr(self, '_data', None)
        shaped_changed = current is not None and current.shape[:self.basedim] != value.shape[:self.basedim]

        # actually set the data
        setattr(self, '_data', value)

        # send signal if the underlying shape has been modified
        if shaped_changed:
            self._shape_changed()

    @property
    def framed_data(self):
        """
        Core data array reshaped to always include the frame dimension, regardless of nframes.
        """
        return conform_ndim(self.data, self.basedim + 1)

    @property
    def nframes(self):
        """
        Number of data frames.
        """
        return self.shape[-1] if self.data.ndim == self.basedim + 1 else 1

    @property
    def shape(self):
        """
        True shape of the internal data array.
        """
        return self.data.shape

    @property
    def baseshape(self):
        """
        Base shape of the data array (excludes the frame dimension).
        """
        return self.data.shape[:self.basedim]

    @property
    def size(self):
        """
        Total number of elements in the data array.
        """
        return self.data.size

    @property
    def dtype(self):
        """
        Data type.
        """
        return self.data.dtype

    def astype(self, dtype, copy=True, order='K'):
        """
        Copy of the array, casted to a specified type.

        Parameters
        ----------
        dtype : np.dtype
            Target datatype.
        copy : bool
            Return copy if array already has matching datatype.
        order : {‘C’, ‘F’, ‘A’, ‘K’}
            Controls the memory layout order of the result. ‘C’ means C order, ‘F’ means
            Fortran order, ‘A’ means ‘F’ order if all the arrays are Fortran contiguous, ‘C’
            order otherwise, and ‘K’ means as close to the order the array elements appear
            in memory as possible.

        Returns
        -------
        arr : FramedArray
            Array with target datatype.
        """
        if dtype == self.dtype and not copy:
            return self
        return self.new(self.data.astype(dtype=dtype, order=order))

    def _shape_changed(self):
        """
        Event hook that is called when the internal data array shape is updated.
        """
        pass

    def save(self, filename, fmt=None):
        """
        Write array to file.

        Parameters
        ----------
        filename : str
            Target filename to write array to.
        fmt : str
            Optional file format to force.
        """
        from surfa.io.framed import save_framed_array
        save_framed_array(self, filename, fmt=fmt)

    def min(self, nonzero=False, frames=False):
        """
        Compute the minimum.

        Parameters
        ----------
        nonzero : bool
            If enabled, only consider nonzero elements. This is ignored if `frames=True`.
        frames : bool
            Compute min along frame axis.

        Returns
        -------
        scalar or FramedArray
            Returns scalar min value, unless `frames` is set to true, in which case a
            new FramedArray is returned.
        """
        if frames:
            return self.new(self.framed_data.min(axis=-1))
        data = self.data
        if nonzero:
            data = data[data.nonzero()]
        return data.min()

    def max(self, frames=False):
        """
        Compute the maximum.

        Parameters
        ----------
        frames : bool
            Compute max along frame axis. 

        Returns
        -------
        scalar or FramedArray
            Returns scalar max value, unless `frames=True`, in which case a
            new FramedArray is returned.
        """
        if frames:
            return self.new(self.framed_data.max(axis=-1))
        return self.data.max()

    def mean(self, nonzero=False, frames=False):
        """
        Compute the mean.

        Parameters
        ----------
        nonzero : bool
            If enabled, only consider nonzero elements.  This is ignored if `frames=True`.
        frames : bool
            Compute mean along frame axis.

        Returns
        -------
        scalar or FramedArray
            Returns scalar mean value, unless `frames` is set to true, in which case a
            new FramedArray is returned.
        """
        if frames:
            return self.new(self.framed_data.mean(axis=-1))
        data = self.data
        if nonzero:
            data = data[data.nonzero()]
        return data.mean()

    def percentile(self, percentiles, method='linear', nonzero=False):
        """
        Compute the q-th percentile of the data.

        Parameters
        ----------
        percentiles : array_like of float
            Percentile or sequence of percentiles to compute, which must be between 0 and 100 inclusive.
        method : str
            Method to use for estimating the percentile. View numpy percentile doc for more information.
        nonzero : bool
            If enabled, only consider nonzero elements.

        Returns
        -------
        percentile : scalar or ndarray
            Computes percentiles.
        """
        data = self.data
        if nonzero:
            data = data[data.nonzero()]
        return np.percentile(data, percentiles, interpolation=method)

    def clip(self, a_min, a_max):
        """
        Clip the array between low and high values.

        Parameters
        ----------
        a_min, a_max : array_like or None
            Minimum and maximum value. If None, clipping is not performed
            on the corresponding edge
        """
        return self.new(np.clip(self.data, a_min, a_max))

    def round(self):
        """
        Round the array to the nearest integer value.
        """
        return self.new(self.data.round())

    def floor(self):
        """
        Floor the array to integer values.
        """
        return self.new(np.floor(self.data))

    def ceil(self):
        """
        Ceil the array to integer values.
        """
        return self.new(np.ceil(self.data))

    def unique(self):
        """
        Find all unique elements of an array, sorted.
        """
        return np.unique(self)

    def onehot(self, mapping, dtype=np.uint32):
        """
        Convert discrete labels to a one-hot encoded probabilistic map.

        Parameters
        ----------
        mapping : array_like of int
            List of label indices to one-hot encode. Label order is preserved.
        dtype : np.dtype
            Output segmentation datatype.

        Returns
        -------
        FramedArray
            Multi-frame one-hot segmentation array.
        """
        if self.nframes > 1:
            raise RuntimeError(f'cannot onehot-encode labels with more than 1 frame, but array has {self.nframes} frames')

        mapping = np.asarray(mapping)
        if mapping.ndim != 1:
            raise ValueError('label mapping must be a 1D list')

        nlabels = len(mapping)
        inttype = np.uint16 if nlabels < np.iinfo(np.uint16).max else np.uint32
        recoder = np.zeros(self.max() + 1, dtype=inttype)
        recoder[mapping] = np.arange(nlabels)

        dsize = self.data.size
        flat = np.zeros((dsize, nlabels), dtype=dtype)
        flat[np.arange(dsize), recoder[self.data.ravel()]] = 1
        flat.shape = (*self.baseshape, nlabels)
        return self.new(flat)

    def collapse(self, mapping=None):
        """
        Collapse a one-hot encoded probabilistic map to discrete labels.

        Parameters
        ----------
        mapping : array_like of int
            List of label indices that correspond to probabilistic data frames encoded.

        Returns
        -------
        FramedArray
            Collapsed, discrete segmentation array.
        """
        if self.nframes == 1:
            raise RuntimeError('cannot collapse probabilities with only 1 frame')
        
        inttype = np.uint16 if self.nframes < np.iinfo(np.uint16).max else np.uint32
        seg = np.zeros(self.baseshape, dtype=inttype)
        np.argmax(self, axis=-1, out=seg)
        
        if mapping is not None:
            mapping = np.asarray(mapping)
            if mapping.ndim != 1:
                raise ValueError('label mapping must be a 1D list')
            seg = mapping[seg]

        return self.new(seg)

    # numpy array wrapping

    def __array__(self):
        return self.data

    # comparison operators

    def __eq__(self, other):
        return self.new(self.data == np.asarray(other))
    
    def __ne__(self, other):
        return self.new(self.data != np.asarray(other))

    def __lt__(self, other):
        return self.new(self.data < np.asarray(other))

    def __le__(self, other):
        return self.new(self.data <= np.asarray(other))

    def __gt__(self, other):
        return self.new(self.data > np.asarray(other))

    def __ge__(self, other):
        return self.new(self.data >= np.asarray(other))

    # unary operators

    def __pos__(self):
        return self.new(+self.data)

    def __neg__(self):
        return self.new(-self.data)

    # binary operators

    def __and__(self, other):
        return self.new(self.data & np.asarray(other))

    def __or__(self, other):
        return self.new(self.data | np.asarray(other))

    def __add__(self, other):
        return self.new(self.data + np.asarray(other))

    def __radd__(self, other):
        return self.new(np.asarray(other) + self.data)

    def __sub__(self, other):
        return self.new(self.data - np.asarray(other))

    def __rsub__(self, other):
        return self.new(np.asarray(other) - self.data)

    def __mul__(self, other):
        return self.new(self.data * np.asarray(other))

    def __rmul__(self, other):
        return self.new(np.asarray(other) * self.data)

    def __truediv__(self, other):
        return self.new(self.data / np.asarray(other))

    def __rtruediv__(self, other):
        return self.new(np.asarray(other) / self.data)

    def __pow__(self, other):
        return self.new(self.data ** np.asarray(other))

    # assignment operators

    def __setitem__(self, key, value):
        if isinstance(key, FramedArray):
            key = np.asarray(key)
        if isinstance(value, FramedArray):
            value = np.asarray(value)
        self.data[key] = value

    def __iadd__(self, other):
        self.data += np.asarray(other)
        return self

    def __isub__(self, other):
        self.data -= np.asarray(other)
        return self

    def __imul__(self, other):
        self.data *= np.asarray(other)
        return self

    def __itruediv__(self, other):
        self.data /= np.asarray(other)
        return self



from .utils import check_file_readability
from .affine import load_affine
from .framed import load_volume
from .framed import load_slice
from .framed import load_overlay
from .labels import load_label_lookup
from .mesh import load_mesh



import os
import shutil
import tempfile
import numpy as np

from surfa import Mesh
from surfa.system import run
from surfa.system import collect_output
from surfa.image import cast_image
from surfa.mesh import cast_overlay


class Freeview:

    def __init__(self, title=None, debug=False):
        """
        A visualization class that wraps the `freeview` command.

        This function assumes FreeView is installed and the `freeview` command is
        accessible in the shell. Images can be configured in the wrapper like so:

            fv = Freeview()
            fv.add_image(img)
            fv.add_image(seg, colormap='lut', opacity=0.5)
            fv.add_mesh(mesh, overlay=overlay)
            fv.show()

        For a quicker but more limited way to wrap freeview, see the `fv()` function.
        """
        self.tempdir = None
        self.debug = debug
        self.title = title
        self.isshown = False
        self.arguments = []

        # first check if freeview is even accessible
        self.fvpath = shutil.which('freeview')
        if self.fvpath is None:
            raise RuntimeError('cannot find `freeview` command in shell')
        if self.debug:
            print(f'using freeview command from {self.fvpath}')

        self.tempdir = tempfile.mkdtemp()
        if self.debug:
            print(f'creating temporary directory at {self.tempdir}')

    def __del__(self):
        """
        If the session is shutting down and the window was never opened, make sure
        to remove the temporary directory.
        """
        if not self.isshown and self.tempdir is not None and os.path.exists(self.tempdir):
            shutil.rmtree(self.tempdir)

    def add_image(self, img, **kwargs):
        """
        Adds an image to the freeview window. Any key/value tags allowed as a `-v` option
        in the freeview command line can be provided as an additional argument.

        Parameters
        ----------
        img: array_like, FramedImage, or str
            Image or filename to load in the freeview session.
        **kwargs:
            Additional options to append as key/values tags to the freeview volume argument.
        """
        # convert the input to a proper file (if it's not one already)
        if isinstance(img, str):
            if not os.path.isfile(img):
                print(f'freeview error: image file {img} does not exist')
                return
            filename = img
        else:
            img = cast_image(img, allow_none=False)
            filename = _unique_filename('image', '.mgz', self.tempdir)
            img.save(filename)
            if self.debug:
                print(f'wrote image to {filename}')

        # configure the corresponding freeview argument
        self.arguments.append('-v ' + filename + _convert_kwargs_to_tags(kwargs))

    def add_mesh(self, mesh, overlay=None, annot=None, **kwargs):
        """
        Adds an image to the freeview window. Any key/value tags allowed as a `-v` option
        in the freeview command line can be provided as an additional argument.

        Parameters
        ----------
        img: array_like, FramedImage, or str
            Image or filename to load in the freeview session.
        overlay : Overlay or sequence of Overlays
            Load overlay on mesh data.
        annot : Overlay or sequence of Overlays
            Load annotation on mesh data.
        **kwargs:
            Additional options to append as key/values tags to the freeview volume argument.
        """
        # convert the input to a proper file (if it's not one already)
        if isinstance(mesh, str):
            if not os.path.isfile(mesh):
                print(f'freeview error: mesh file {mesh} does not exist')
                return
            mesh_filename = mesh
        elif isinstance(mesh, Mesh):
            mesh_filename = _unique_filename('mesh', '', self.tempdir)
            mesh.save(mesh_filename)
            if self.debug:
                print(f'wrote mesh to {mesh_filename}')
        else:
            raise ValueError(f'expected type Mesh to add_mesh, but got type {mesh.__class__.__name__}')

        # extra tags for the mesh
        tags = ''

        # configure any overlays
        if overlay is not None:
            overlay = [overlay] if not isinstance(overlay, (list, tuple)) else overlay
            for c in overlay:
                c = FreeviewOverlay(c) if not isinstance(c, FreeviewOverlay) else c
                filename = _unique_filename(c.name, '.mgz', self.tempdir)
                c.arr.save(filename)
                if self.debug:
                    print(f'wrote overlay to {filename}')
                tags += f':overlay={filename}' + c.tags()

        # configure any annotations
        if annot is not None:
            annot = [annot] if not isinstance(annot, (list, tuple)) else annot
            for c in annot:
                c = FreeviewAnnot(c) if not isinstance(c, FreeviewAnnot) else c
                filename = _unique_filename(c.name, '.annot', self.tempdir)
                c.arr.save(filename)
                if self.debug:
                    print(f'wrote annotation to {filename}')
                tags += f':annot={filename}'

        # configure the corresponding freeview argument
        self.arguments.append('-f ' + mesh_filename + tags + _convert_kwargs_to_tags(kwargs))

    def show(self, background=True, threads=None):
        """
        Opens the configured FreeView window.

        Parameters
        ----------
            background : bool
                Run FreeView window as a background process.
            threads : int
                Number of OMP threads available to FreeView.
        """

        # compile the command
        command = self.fvpath + ' ' + ' '.join(self.arguments)

        # add window title
        if self.title is not None:
            title = self.title.replace('"', "'")
            command = f'{command} -subtitle "{title}"'

        # be sure to remove the temporary directory (if it exists) after freeview closes
        command = f'{command} ; rm -rf {self.tempdir}'

        # freeview can be buggy when run remotely, so let's test if VGL is
        # available to wrap the process
        vgl = _find_vgl()
        if vgl is not None:
            command = f'{vgl} {command}'

        # set number of OMP threads if provided
        if threads is not None:
            command = f'OMP_NUM_THREADS={threads} {command}'

        if self.debug:
            print('running FreeView command:')
            print(command)

        # mark the window has been opened to avoid future modification
        self.isshown = True

        # run it
        run(command, background=background)


class FreeviewOverlay:

    def __init__(self, arr, name='overlay', threshold=None, opacity=None):
        """
        Configuration for freeview overlays.
        """
        self.arr = cast_overlay(arr, allow_none=False)
        self.name = name
        self.threshold = threshold
        self.opacity = opacity

    def tags(self):
        tags = ''
        tags += '' if self.threshold is None else f':overlay_threshold=' + ','.join(str(x) for x in config.threshold)
        tags += '' if self.opacity is None else f':overlay_opacity={self.opacity}'
        return tags


class FreeviewAnnot:

    def __init__(self, arr, name='annotation'):
        """
        Configuration for freeview annotations.
        """
        self.arr = cast_overlay(arr, allow_none=False)
        self.name = name


def fv(*args, **kwargs):
    """
    Freeview wrapper to quickly load an arbitray number of elements. Inputs
    can be existing filenames, images, meshes, or numpy arrays. Lists
    are also supported. Use the `Freeview` class directly to configure a
    more advanced session.

    Parameters
    ----------
    *args : array_like, FramedImage, Mesh, or str
        Elements to load in FreeView window.
    **kwargs
        Parameters forwarded to the Freeview constructor.
    """
    background = kwargs.pop('background', True)

    # initialize session
    fv = Freeview(**kwargs)

    # expand any nested lists/tuples within args
    def flatten(deep):
        for el in deep:
            if isinstance(el, (list, tuple)):
                yield from flatten(el)
            else:
                yield el

    # cycle through arguments
    for arg in flatten(args):
        if isinstance(arg, Mesh):
            fv.add_mesh(arg)
        else:
            fv.add_image(arg)

    # show the window
    fv.show(background=background)


def _find_vgl():
    """
    Locate the VGL wrapper if installed.
    """
    have_key = os.path.isfile('/etc/opt/VirtualGL/vgl_xauth_key')
    vgl_path = shutil.which('vglrun')
    if vgl_path is None:
        vgl_path = shutil.which('vglrun', path='/usr/pubsw/bin')
    if vgl_path is None:
        return None
    islocal = any([os.environ.get('DISPLAY', '').endswith(string) for string in (':0', ':0.0')])
    no_glx = 'NV-GLX' in collect_output('xdpyinfo')[0]
    if not islocal and not nvglx:
        return vgl_path
    return None


def _convert_kwargs_to_tags(kwargs):
    """
    Converts a kwargs dictionary to freeview key/value tags
    """
    tags = ''
    for key, value in kwargs.items():
        if isinstance(value, (list, tuple)):
            value = ','.join(str(x) for x in value)
        if value is not None:
            vale = value.replace(' ', '-')
            tags += f':{key}={value}'
    return tags


def _unique_filename(filename, extension, directory):
    """
    Identifies a unique filename from a base string in a directory.
    """
    # make sure extension start with a dot
    if extension and not extension.startswith('.'):
        extension = f'.{extension}'

    # check if it's unique
    fullpath = os.path.join(directory, f'{filename}{extension}')
    if not os.path.exists(fullpath):
        return fullpath

    # append numbers until a unique filename is created (stop after 10k tries)
    for n in range(2, 10000):
        fullpath = os.path.join(directory, f'{filename}-{n:02d}{extension}')
        if not os.path.exists(fullpath):
            return fullpath
    raise RuntimeError(f'could not generate a unique filename for {filename} after trying many times')



import os
from copy import deepcopy
import numpy as np

from surfa.core.array import check_array
from surfa.transform import ImageGeometry
from surfa.transform import image_geometry_equal
from surfa.transform import cast_image_geometry
from surfa.transform import cast_space


class Mesh:

    def __init__(self, vertices, faces=None, space='surf', geometry=None, metadata=None):
        """
        Triangular mesh topology represented by arrays of vertices and faces.

        Parameters
        ----------
        vertices : (V, 3) float
            TODO
        faces : (F, 3) int
            TODO
        space : Space
            Coordinate space of the point data. Defaults to the 'surface' coordinate system.
        geometry : ImageGeometry
            Geometry mapping the point data to world and image coordinates.
        metadata : dict
            Dictionary containing arbitrary array metadata.
        """
        self.vertices = vertices
        self.faces = faces
        self.space = space
        self.geom = geometry

        # initialize and set the private metadata dictionary
        self._metadata = {}
        self.metadata = metadata

    def copy(self):
        """
        Return a deep copy of the object.
        """
        return deepcopy(self)

    def save(self, filename, fmt=None):
        """
        Write mesh to file.

        Parameters
        ----------
        filename : str
            Target filename to write array to.
        fmt : str
            Optional file format to force.
        """
        from surfa.io.mesh import save_mesh
        save_mesh(self, filename, fmt=fmt)

    @property
    def vertices(self):
        """
        Mesh 3D point positions defined by a (V, 3) array, where V corresponds to
        the number of vertices.
        """
        return self._vertices

    @vertices.setter
    def vertices(self, vertices):
        vertices = np.asarray(vertices, dtype=np.float64)
        check_array(vertices, ndim=2, name='vertices')
        if vertices.shape[-1] != 3:
            raise ValueError(f'expected shape (V, 3) for vertices array, but got {vertices.shape}')
        self._vertices = vertices

    @property
    def nvertices(self):
        """
        Total number of vertices in the mesh.
        """
        return len(self.vertices)

    @property
    def faces(self):
        """
        Mesh triangle faces defined by a (F, 3) array, where F corresponds to the
        total number of mesh faces.
        """
        return self._faces

    @faces.setter
    def faces(self, faces):
        faces = np.asarray(faces, dtype=np.int64)
        check_array(faces, ndim=2, name='faces')
        if faces.shape[-1] != 3:
            raise ValueError(f'expected shape (F, 3) for faces array, but got {faces.shape}')
        self._faces = faces

    @property
    def nfaces(self):
        """
        Total number of faces in the mesh.
        """
        return len(self.faces)

    @property
    def space(self):
        """
        Coordinate space of the points.
        """
        return self._space
    
    @space.setter
    def space(self, value):
        self._space = cast_space(value, allow_none=False, copy=True)

    @property
    def geom(self):
        """
        Geometry that maps mesh coordinates to image and world spaces.
        """
        return self._geometry

    @geom.setter
    def geom(self, geometry):
        if geometry is None:
            geometry = ImageGeometry(shape=(256, 256, 256))
        else:
            geometry = cast_image_geometry(geometry, copy=True)
        setattr(self, '_geometry', geometry)

    @property
    def metadata(self):
        """
        Metadata dictionary.
        """
        return self._metadata

    @metadata.setter
    def metadata(self, value):
        """
        Replace the metadata dictionary. Will always make a deep copy of the new dictionary.
        """
        self._metadata = deepcopy(value) if value is not None else {}

    def convert(self, space=None, geometry=None, copy=True):
        """
        Converts mesh vertex positions into a new coordinate space, given a particular
        image geometry.

        Parameters
        ----------
        space : Space
            Target coordinate space.
        geometry : Geometry
            Target image geometry.
        copy : bool
            Return a copy of the mesh when target space and geometry conditions
            are already satisfied.s

        Returns
        -------
        Mesh
            Mesh with converted vertex positions.
        """
        space = self.space if space is None else cast_space(space)
        geometry = self.geom if geometry is None else cast_image_geometry(geometry)

        same_geom = image_geometry_equal(geometry, self.geom)
        same_space = space == self.space

        # return self if no changes are necessary
        if same_geom and same_space:
            return self.copy() if copy else self

        # make copy of the mesh
        converted = self.copy()

        if same_geom:
            # only need to transform the points once if the geometry doesn't change
            converted.vertices = self.geom.affine(self.space, space).transform(self.vertices)
            converted.space = space
        else:
            # if we're updating the geometry, we'll need to pass through world space first
            aff = geometry.affine('world', space) @ self.geom.affine(self.space, 'world')
            converted.vertices = aff(self.vertices)
            converted.space = space
            converted.geom = geometry

        return converted



import os
import numpy as np

from surfa.io import protocol
from surfa.core.labels import LabelLookup
from surfa.io import check_file_readability


def load_label_lookup(filename, fmt=None):
    """
    Load a `LabelLookup` from file.

    Parameters
    ----------
    filename : str
        File path to read.
    fmt : str, optional
        Forced file format. If None (default), file format is extrapolated
        from extension.

    Returns
    -------
    LabelLookup
        Loaded label lookup table. 
    """
    check_file_readability(filename)

    if fmt is None:
        iop = protocol.find_protocol_by_extension(labels_io_protocols, filename)
        if iop is None:
            raise ValueError(f'cannot determine file format from extension for {filename}')
    else:
        iop = protocol.find_protocol_by_name(labels_io_protocols, fmt)
        if iop is None:
            raise ValueError(f'unknown file format {fmt}')

    return iop().load(filename)


def save_label_lookup(labels, filename, fmt=None):
    """
    Save a `LabelLookup` object to file.

    Parameters
    ----------
    labels : LabelLookup
        Object to write.
    filename: str
        Destination file path.
    fmt : str
        Forced file format. If None (default), file format is extrapolated
        from extension.
    """
    if fmt is None:
        iop = protocol.find_protocol_by_extension(labels_io_protocols, filename)
        if iop is None:
            raise ValueError(f'cannot determine file format from extension for {filename}')
    else:
        iop = protocol.find_protocol_by_name(labels_io_protocols, fmt)
        if iop is None:
            raise ValueError(f'unknown file format {fmt}')
        filename = iop.enforce_extension(filename)

    iop().save(labels, filename)


class FreeSurferLookupTableIO(protocol.IOProtocol):
    """
    LabelLookup IO protocol for FreeSurfer LUT files.
    """

    name = 'ctab'
    extensions = ('.ctab', '.lut', '.txt')

    def load(self, filename):
        """
        Read LabelLookup from an LUT file.

        Parameters
        ----------
        filename : str
            File path to read.

        Returns
        -------
        LabelLookup
            Object loaded from file.
        """
        labels = LabelLookup()
        with open(filename, 'r') as file:
            lines = file.readlines()
        for line in lines:
            split = line.lstrip().split()
            if split and not split[0].startswith('#'):
                index, name = split[:2]
                if len(split) >= 5:
                    color = np.asarray(list(map(int, split[2:6])), dtype=np.float64)
                    color[3] = (255 - color[3]) / 255  # invert alpha value
                else:
                    color = None
                labels[int(index)] = (name, color)
        return labels

    def save(self, labels, filename):
        """
        Write LabelLookup to an LTA file.

        Parameters
        ----------
        labels : LabelLookup
            Label lookup table to save.
        filename : str
            Target file path.
        """
        with open(filename, 'w') as file:
            col1 = len(str(max(labels.keys()))) + 1  # find largest index
            col2 = max([len(elt.name) for elt in labels.values()]) + 2  # find longest name
            file.write('#'.ljust(col1) + 'Label Name'.ljust(col2) + '  R    G    B    A\n\n')
            for index, elt in labels.items():
                color = elt.color.copy()
                color[3] = 255 * (1 - color[3])  # invert alpha value
                colorstr = ' '.join([str(c).rjust(4) for c in color.astype(np.uint8)])
                file.write(str(index).ljust(col1) + elt.name.ljust(col2) + colorstr + '\n')


# enabled IO protocol classes
labels_io_protocols = [
    FreeSurferLookupTableIO,
]
